/**
 * @package webpack-dev-service
 * @license MIT
 * @version 0.15.7
 * @author nuintun <nuintun@qq.com>
 * @description A koa 2 middleware for webpack development and hot reloading.
 * @see https://github.com/nuintun/webpack-dev-service#readme
 */

'use strict';

const events = require('./events.cjs');
const Overlay = require('./ui/Overlay.cjs');
const Progress = require('./ui/Progress.cjs');
const hot = require('./hot.cjs');

/**
 * @module client
 */
function createClient(options) {
  let updateTimer;
  const UPDATE_DELAY = 128;
  const RETRY_INTERVAL = 3000;
  const progress = new Progress.Progress();
  const overlay = new Overlay.Overlay(options.name);
  const fallback = () => {
    if (options.reload) {
      self.location.reload();
    } else {
      console.warn('Use fallback update but you turn off live reload, please reload by yourself.');
    }
  };
  const applyUpdateAsync = () => {
    updateTimer = self.setTimeout(() => {
      hot.applyUpdate(options.hmr, fallback);
    }, UPDATE_DELAY);
  };
  const onInvalid = () => {
    clearTimeout(updateTimer);
    if (options.progress) {
      progress.update(0);
      progress.show();
    }
  };
  const onProgress = ({ percentage }) => {
    if (options.progress) {
      progress.update(percentage);
    }
  };
  const onHash = ({ hash }) => {
    hot.updateHash(hash);
  };
  const setIssues = (type, issues) => {
    if (options.overlay) {
      overlay.setIssues(type, issues);
    }
    const maps = {
      errors: ['\x1b[38;2;100;30;22m\x1b[48;2;255;95;88m ERROR \x1b[0m', 'error'],
      warnings: ['\x1b[38;2;32;39;35m\x1b[48;2;255;189;46m WARNING \x1b[0m', 'warn']
    };
    const [name, method] = maps[type];
    const debug = console[method];
    for (const { moduleName, chunkName, message } of issues) {
      const filename = moduleName || chunkName || 'unknown';
      debug(`${name} in ${filename}\n${message}`);
    }
  };
  const onIssues = ({ errors, warnings }) => {
    progress.update(1);
    progress.hide();
    setIssues('errors', errors);
    setIssues('warnings', warnings);
    if (options.overlay) {
      overlay.show();
    }
    if (errors.length <= 0) {
      applyUpdateAsync();
    }
  };
  const onOk = () => {
    progress.update(1);
    progress.hide();
    overlay.hide();
    applyUpdateAsync();
  };
  const parseMessage = message => {
    try {
      return JSON.parse(message.data);
    } catch {
      return null;
    }
  };
  const createWebSocket = url => {
    const ws = new WebSocket(url);
    ws.onmessage = message => {
      const parsed = parseMessage(message);
      if (parsed) {
        const { action, payload } = parsed;
        switch (action) {
          case 'invalid':
            onInvalid();
            break;
          case 'progress':
            onProgress(payload);
            break;
          case 'hash':
            onHash(payload);
            break;
          case 'issues':
            onIssues(payload);
            break;
          case 'ok':
            onOk();
            break;
        }
        events.emit(action, payload, options);
      }
    };
    ws.onclose = event => {
      overlay.hide();
      progress.hide();
      switch (event.code) {
        case 4000:
        case 4001:
          fallback();
          break;
        default:
          if (!event.wasClean) {
            setTimeout(() => {
              createWebSocket(url);
            }, RETRY_INTERVAL);
          }
          break;
      }
    };
  };
  const { uuid } = options;
  const input = new URL(`${options.origin}${options.path}`);
  if (uuid) {
    input.searchParams.set('uuid', uuid);
  }
  createWebSocket(input.href);
}

exports.createClient = createClient;
