/**
 * @package webpack-dev-service
 * @license MIT
 * @version 0.15.7
 * @author nuintun <nuintun@qq.com>
 * @description A koa 2 middleware for webpack development and hot reloading.
 * @see https://github.com/nuintun/webpack-dev-service#readme
 */

'use strict';

/**
 * @module hot
 */
// Last error.
let error;
// Last update hash.
let hash = __webpack_hash__;
// HMR status.
let status = 'idle';
/**
 * @function isUpdateAvailable
 * @description Is there a newer version of this code available.
 */
function isUpdateAvailable() {
  // __webpack_hash__ is the hash of the current compilation.
  // It's a global variable injected by webpack.
  return hash !== __webpack_hash__;
}
/**
 * @function updateStatus
 * @description Update hot status.
 * @param value The new status of the hot update.
 */
function updateStatus(value) {
  status = value;
}
// Initialize status.
if (module.hot) {
  updateStatus(module.hot.status());
}
/**
 * @function updateHash
 * @description Update webpack hash.
 * @param value The new hash value.
 */
function updateHash(value) {
  hash = value;
}
/**
 * @function applyUpdate
 * @description Apply update.
 * @param hmr Whether to enable HMR.
 * @param fallback Fallback function when HMR fail.
 */
function applyUpdate(hmr, fallback) {
  // Update available.
  if (isUpdateAvailable()) {
    // HMR enabled.
    if (hmr && module.hot) {
      switch (status) {
        case 'idle':
          // Update status.
          updateStatus('check');
          // Auto check and apply updates.
          module.hot
            .check(true)
            .then(updated => {
              // Update status.
              updateStatus(module.hot.status());
              // When updated modules is available,
              // it indicates server is ready to serve new bundle.
              if (updated) {
                // While update completed, do it again until no update available.
                applyUpdate(hmr, fallback);
              }
            })
            .catch(exception => {
              // Get status.
              const status = module.hot.status();
              // Update status.
              switch (status) {
                case 'fail':
                case 'abort':
                  updateStatus(status);
                  break;
                default:
                  updateStatus('fail');
                  break;
              }
              // Cache error.
              error = exception;
              // Call fallback.
              fallback(error);
            });
          break;
        case 'fail':
        case 'abort':
          // Call fallback.
          fallback(error);
          break;
      }
    } else {
      // HMR disabled.
      fallback();
    }
  }
}

exports.applyUpdate = applyUpdate;
exports.updateHash = updateHash;
