/**
 * @package webpack-dev-service
 * @license MIT
 * @version 0.15.7
 * @author nuintun <nuintun@qq.com>
 * @description A koa 2 middleware for webpack development and hot reloading.
 * @see https://github.com/nuintun/webpack-dev-service#readme
 */

'use strict';

const createETag = require('etag');
const ReadStream = require('./ReadStream.cjs');
const utils = require('../utils.cjs');
const fs = require('./utils/fs.cjs');
const node_path = require('node:path');
const path = require('./utils/path.cjs');
const http = require('./utils/http.cjs');

function _interopDefault(e) {
  return e && e.__esModule ? e : { default: e };
}

const createETag__default = /*#__PURE__*/ _interopDefault(createETag);

/**
 * @module Service
 */
/**
 * @class Service
 */
class Service {
  #root;
  #options;
  /**
   * @constructor
   * @description Create file service.
   * @param root The file service root.
   * @param options The file service options.
   */
  constructor(root, options) {
    this.#options = options;
    this.#root = path.unixify(node_path.resolve(root));
  }
  /**
   * @private
   * @method #isIgnore
   * @description Check if path is ignore.
   * @param path The path to check.
   */
  #isIgnore(path) {
    const { ignore } = this.#options;
    return (utils.isFunction(ignore) ? ignore(path) : false) === true;
  }
  /**
   * @private
   * @method #setupHeaders
   * @description Setup headers.
   * @param context The koa context.
   * @param path The file path.
   * @param stats The file stats.
   */
  async #setupHeaders({ response }, path, stats) {
    const options = this.#options;
    const { headers } = options;
    // Set status.
    response.status = 200;
    // Set Content-Type.
    response.type = node_path.extname(path);
    // Set headers.
    if (headers) {
      if (utils.isFunction(headers)) {
        const fields = await headers(path, stats);
        if (fields) {
          response.set(fields);
        }
      } else {
        response.set(headers);
      }
    }
    // Accept-Ranges.
    if (options.acceptRanges === false) {
      // Set Accept-Ranges to none tell client not support.
      response.set('Accept-Ranges', 'none');
    } else {
      // Set Accept-Ranges.
      response.set('Accept-Ranges', 'bytes');
    }
    // ETag.
    if (options.etag === false) {
      // Remove ETag.
      response.remove('ETag');
    } else if (!response.get('ETag')) {
      // Set weak ETag.
      response.set('ETag', createETag__default.default(stats));
    }
    // Last-Modified.
    if (options.lastModified === false) {
      // Remove Last-Modified.
      response.remove('Last-Modified');
    } else if (!response.get('Last-Modified')) {
      // Set last modified from mtime.
      response.set('Last-Modified', stats.mtime.toUTCString());
    }
  }
  /**
   * @public
   * @method respond
   * @description Respond file.
   * @param pathname The pathname.
   * @param context The koa context.
   * @param publicPath The public path.
   */
  async respond(pathname, context, publicPath) {
    // Check public path.
    if (!pathname.startsWith(publicPath)) {
      return false;
    }
    // Get root path.
    const root = this.#root;
    // Slice length.
    const { length } = publicPath;
    // Real pathname.
    const realpath = pathname.slice(length);
    // Get path of file.
    const path$1 = path.unixify(node_path.join(root, realpath));
    // Malicious path (403).
    if (path.isOutRoot(path$1, root)) {
      return false;
    }
    // Is ignore path or file (403).
    if (this.#isIgnore(path$1)) {
      return false;
    }
    // Get options.
    const options = this.#options;
    // File stats.
    const stats = await fs.stat(options.fs, path$1);
    // Check file stats.
    if (
      // File not exist (404 | 500).
      stats == null ||
      // Is directory (403).
      stats.isDirectory() ||
      // Not a directory but has trailing slash (404).
      path.hasTrailingSlash(path$1)
    ) {
      return false;
    }
    // Koa request and response.
    const { request, response } = context;
    // Setup headers.
    await this.#setupHeaders(context, path$1, stats);
    // Conditional get support.
    if (http.isConditionalGET(context)) {
      // Request precondition failure.
      if (http.isPreconditionFailed(context)) {
        return context.throw(412);
      }
      // Request fresh (304).
      if (request.fresh) {
        // Set status.
        response.status = 304;
        // Set body null.
        response.body = null;
        // File found.
        return true;
      }
    }
    // Head request.
    if (request.method === 'HEAD') {
      // Set Content-Length.
      response.length = stats.size;
      // Set body null
      response.body = null;
      // File found.
      return true;
    }
    // Parsed ranges.
    const ranges = http.parseRanges(context, stats);
    // 416
    if (ranges === -1) {
      // Set Content-Range.
      response.set('Content-Range', `bytes */${stats.size}`);
      // Unsatisfiable 416.
      return context.throw(416);
    }
    // 400.
    if (ranges === -2) {
      return context.throw(400);
    }
    // Set response body.
    response.body = new ReadStream.ReadStream(path$1, ranges, options);
    // File found.
    return true;
  }
}

exports.Service = Service;
