/**
 * @package webpack-dev-service
 * @license MIT
 * @version 0.15.7
 * @author nuintun <nuintun@qq.com>
 * @description A koa 2 middleware for webpack development and hot reloading.
 * @see https://github.com/nuintun/webpack-dev-service#readme
 */

'use strict';

const Service = require('./Service.cjs');
const ready = require('./utils/ready.cjs');
const http = require('./utils/http.cjs');
const paths = require('./utils/paths.cjs');

/**
 * @module middleware
 */
function getFileServices(context, stats) {
  const cached = context.services;
  // Cache hit.
  if (cached) {
    return cached;
  }
  const { options } = context;
  const paths$1 = paths.getPaths(stats);
  const services = [];
  // Get the file services.
  for (const [outputPath, publicPath] of paths$1) {
    services.push([publicPath, new Service.Service(outputPath, options)]);
  }
  // Cache services.
  context.services = services;
  // Return services.
  return services;
}
function getFileServicesAsync(context) {
  return new Promise(resolve => {
    const { stats } = context;
    // If stats exists, resolve immediately.
    if (stats) {
      resolve(getFileServices(context, stats));
    } else {
      // Otherwise, wait until bundle finished.
      ready.ready(context, stats => {
        resolve(getFileServices(context, stats));
      });
    }
  });
}
function middleware(context) {
  // Middleware.
  return async (ctx, next) => {
    const { request } = ctx;
    const pathname = http.decodeURI(request.path);
    // Pathname decode failed or includes null byte(s).
    if (pathname === -1 || pathname.includes('\0')) {
      return ctx.throw(400);
    }
    // Get request method.
    const { method } = request;
    // Only support GET and HEAD (405).
    if (method === 'GET' || method === 'HEAD') {
      // Get the file services.
      const services = await getFileServicesAsync(context);
      // Try to respond.
      for (const [publicPath, service] of services) {
        if (await service.respond(pathname, ctx, publicPath)) {
          return;
        }
      }
    }
    // Not respond.
    await next();
  };
}

exports.middleware = middleware;
