/**
 * @package webpack-dev-service
 * @license MIT
 * @version 0.15.7
 * @author nuintun <nuintun@qq.com>
 * @description A koa 2 middleware for webpack development and hot reloading.
 * @see https://github.com/nuintun/webpack-dev-service#readme
 */

'use strict';

const webpack = require('webpack');
const uuid = require('uuid');
const node_buffer = require('node:buffer');
const node_path = require('node:path');
const node_url = require('node:url');
const WebSocket = require('ws');
const utils$1 = require('../utils.cjs');
const utils = require('./utils.cjs');

function _interopDefault(e) {
  return e && e.__esModule ? e : { default: e };
}

const webpack__default = /*#__PURE__*/ _interopDefault(webpack);
const WebSocket__default = /*#__PURE__*/ _interopDefault(WebSocket);

/**
 * @module Socket
 */
function entrypoint() {
  const filename = require.main.filename;
  return node_path.dirname(filename);
}
const client = node_path.resolve(entrypoint(), '../../client/main.cjs');
class Socket {
  // Readonly props.
  #logger;
  #server;
  #options;
  constructor(compiler, options) {
    this.#options = utils.getOptions(options);
    this.#logger = compiler.getInfrastructureLogger(utils$1.PLUGIN_NAME);
    this.#server = new WebSocket.WebSocketServer({ noServer: true, path: this.#options.path });
    const compilers = utils$1.getCompilers(compiler);
    const contexts = new Map();
    for (const compiler of compilers) {
      const uuid$1 = uuid.v7();
      const context = {
        uuid: uuid$1,
        stats: null,
        percentage: -1,
        clients: new Set()
      };
      contexts.set(uuid$1, context);
      this.#setupHooks(compiler, context);
      this.#setupPlugins(compiler, context);
    }
    this.#setupWss(contexts);
  }
  #setupHooks(compiler, context) {
    const { hooks } = compiler;
    const statsOptions = utils.getStatsOptions(compiler);
    hooks.invalid.tap(utils$1.PLUGIN_NAME, (path, timestamp) => {
      // Set stats to null.
      context.stats = null;
      // Reset percentage.
      context.percentage = -1;
      // Broadcast invalid.
      this.#broadcast(context.clients, 'invalid', { path, timestamp });
    });
    hooks.done.tap(utils$1.PLUGIN_NAME, stats => {
      // Get json stats.
      const jsonStats = stats.toJson(statsOptions);
      // Hack builtAt.
      if (jsonStats.builtAt == null) {
        jsonStats.builtAt = utils.getTimestamp(jsonStats);
      }
      // Cache stats.
      context.stats = jsonStats;
      // Do the stuff in nextTick, because bundle may be invalidated if a change happened while compiling.
      process.nextTick(() => {
        const { stats } = context;
        // Broadcast stats.
        if (stats) {
          this.#broadcastStats(context.clients, stats);
        }
      });
    });
  }
  #setupPlugins(compiler, context) {
    const options = this.#options;
    const params = new URLSearchParams();
    params.set('uuid', context.uuid);
    params.set('path', options.path);
    const { wss } = options;
    if (wss != null) {
      params.set('wss', wss ? 'true' : 'false');
    }
    params.set('name', compiler.name || 'rspack');
    params.set('hmr', options.hmr ? 'true' : 'false');
    params.set('reload', options.reload ? 'true' : 'false');
    params.set('overlay', options.overlay ? 'true' : 'false');
    params.set('progress', options.progress ? 'true' : 'false');
    const plugins = [
      new webpack__default.default.HotModuleReplacementPlugin(),
      // Auto add hot client to entry.
      new webpack__default.default.EntryPlugin(compiler.context, `${client}?${params}`, {
        // Don't create runtime.
        runtime: false
      })
    ];
    if (options.progress) {
      plugins.push(
        new webpack__default.default.ProgressPlugin((percentage, status, ...messages) => {
          if (percentage > context.percentage) {
            context.percentage = percentage;
            this.#broadcast(context.clients, 'progress', { status, messages, percentage });
          }
        })
      );
    }
    for (const plugin of plugins) {
      plugin.apply(compiler);
    }
  }
  #setupWss(contexts) {
    const logger = this.#logger;
    const server = this.#server;
    const { path } = this.#options;
    server.on('connection', (client, { url: input = path }) => {
      const url = new node_url.URL(input, utils.BASE_URL);
      const uuid = url.searchParams.get('uuid');
      if (uuid) {
        const context = contexts.get(uuid);
        if (context) {
          const { stats } = context;
          context.clients.add(client);
          logger.log('client connected');
          client.on('close', () => {
            context.clients.delete(client);
            logger.log('client disconnected');
          });
          if (stats) {
            const clients = new Set([client]);
            this.#broadcastStats(clients, stats);
          }
        } else {
          client.close(4001, 'compiler not found');
        }
      } else {
        client.close(4000, 'uuid not found');
      }
    });
    server.on('error', error => {
      logger.error(error);
    });
  }
  #broadcast(clients, action, payload) {
    for (const client of clients) {
      if (client.readyState === WebSocket__default.default.OPEN) {
        client.send(JSON.stringify({ action, payload }));
      }
    }
  }
  #broadcastStats(clients, stats) {
    if (clients.size > 0) {
      const { hash, errors, warnings, builtAt: timestamp } = stats;
      this.#broadcast(clients, 'hash', { hash, timestamp });
      if (utils.hasIssues(errors) || utils.hasIssues(warnings)) {
        this.#broadcast(clients, 'issues', { errors, warnings, timestamp });
      } else {
        this.#broadcast(clients, 'ok', { timestamp });
      }
    }
  }
  upgrade(context) {
    const server = this.#server;
    const { req: request } = context;
    if (utils.isUpgradable(context) && server.shouldHandle(request)) {
      context.respond = false;
      const { socket } = context;
      const head = node_buffer.Buffer.alloc(0);
      server.handleUpgrade(request, socket, head, client => {
        server.emit('connection', client, request);
      });
      return true;
    }
    return false;
  }
}

exports.Socket = Socket;
