/**
 * @package webpack-dev-service
 * @license MIT
 * @version 0.15.7
 * @author nuintun <nuintun@qq.com>
 * @description A koa 2 middleware for webpack development and hot reloading.
 * @see https://github.com/nuintun/webpack-dev-service#readme
 */

import { Ansi } from '@nuintun/ansi';

/**
 * @module utils
 */
const ansi = new Ansi();
function escapeHTML(text) {
  return text.replace(/[&<>"']/gm, match => {
    switch (match) {
      case '&':
        return '&amp;';
      case '<':
        return '&lt;';
      case '>':
        return '&gt;';
      case '"':
        return '&quot;';
      case "'":
        return '&#x27;';
      default:
        return match;
    }
  });
}
function blockToHTML({ style, value, url }) {
  const styles = [];
  const textDecorations = [];
  if (style.dim) {
    styles.push(`opacity: 0.5`);
  }
  if (style.bold) {
    styles.push(`font-weight: bold`);
  }
  if (style.italic) {
    styles.push(`font-style: italic`);
  }
  if (style.inverse) {
    styles.push(`filter: invert(1)`);
  }
  if (style.hidden) {
    styles.push(`visibility: hidden`);
  }
  if (style.blink) {
    textDecorations.push('blink');
  }
  if (style.overline) {
    textDecorations.push('overline');
  }
  if (style.underline) {
    textDecorations.push('underline');
  }
  if (style.strikethrough) {
    textDecorations.push('line-through');
  }
  const { color, background } = style;
  if (color) {
    styles.push(`color: rgb(${color})`);
  }
  if (background) {
    styles.push(`background-color: rgb(${background})`);
  }
  if (textDecorations.length > 0) {
    styles.push(`text-decoration: ${textDecorations.join(' ')}`);
  }
  const escapedValue = escapeHTML(value);
  const href = url ? JSON.stringify(new URL(url).toString()) : null;
  if (styles.length <= 0) {
    if (!href) {
      return escapedValue;
    }
    return `<a href=${href} target="_blank">${escapedValue}</a>`;
  }
  const inlineStyle = JSON.stringify(`${styles.join('; ')};`);
  if (!href) {
    return `<span style=${inlineStyle}>${escapedValue}</span>`;
  }
  return `<a style=${inlineStyle} href=${href} target="_blank">${escapedValue}</a>`;
}
function ansiToHTML(text) {
  let html = '';
  ansi.write(text, block => {
    html += blockToHTML(block);
  });
  ansi.flush(block => {
    html += blockToHTML(block);
  });
  return html;
}
function getRootElement(tagName) {
  const stage = document.createElement(tagName);
  const root = stage.attachShadow({ mode: 'closed' });
  document.body.appendChild(stage);
  return root;
}
function injectCSS(css, root = document.body, styleElement = document.createElement('style')) {
  styleElement.appendChild(document.createTextNode(css.trim()));
  if (!root.contains(styleElement)) {
    root.appendChild(styleElement);
  }
  return styleElement;
}
function appendHTML(html, root = document.body) {
  const nodes = [];
  const parser = new DOMParser();
  const fragment = document.createDocumentFragment();
  const { body } = parser.parseFromString(html.trim(), 'text/html');
  while (body.firstChild) {
    nodes.push(fragment.appendChild(body.firstChild));
  }
  root.appendChild(fragment);
  return nodes;
}

export { ansiToHTML, appendHTML, blockToHTML, escapeHTML, getRootElement, injectCSS };
