/**
 * @package webpack-dev-service
 * @license MIT
 * @version 0.15.7
 * @author nuintun <nuintun@qq.com>
 * @description A koa 2 middleware for webpack development and hot reloading.
 * @see https://github.com/nuintun/webpack-dev-service#readme
 */

import { Buffer } from 'node:buffer';
import { Readable } from 'node:stream';

/**
 * @module ReadStream
 */
const DISPOSE_EVENT = Symbol('dispose');
class ReadStream extends Readable {
  #fs;
  #path;
  #ranges;
  #bytesRead = 0;
  #fd = null;
  #reading = false;
  #currentRangeIndex = 0;
  #readState = 0 /* ReadState.PREFIX */;
  /**
   * @constructor
   * @param path The file path.
   * @param ranges The ranges to read.
   * @param options The stream options.
   */
  constructor(path, ranges, options) {
    const { fs, highWaterMark } = options;
    super({ highWaterMark });
    this.#fs = fs;
    this.#path = path;
    this.#ranges = ranges;
  }
  /**
   * @override
   * @method _construct
   * @param callback The callback.
   */
  _construct(callback) {
    this.#fs.open(this.#path, 'r', (openError, fd) => {
      if (openError == null) {
        this.#fd = fd;
      }
      callback(openError);
    });
  }
  /**
   * @private
   * @method #getRange
   */
  #getRange() {
    return this.#ranges[this.#currentRangeIndex];
  }
  /**
   * @private
   * @method #getPadding
   * @param range The current range.
   */
  #getPadding(range) {
    switch (this.#readState) {
      case 0 /* ReadState.PREFIX */:
        return range.prefix;
      case 2 /* ReadState.SUFFIX */:
        return range.suffix;
    }
  }
  /**
   * @private
   * @method #readFilePadding
   * @param fd The file descriptor.
   * @param range The current range.
   * @param size The number of bytes to read.
   */
  #readFilePadding(fd, range, size) {
    let bytesRead = 0;
    const padding = this.#getPadding(range);
    const hasRangePadding = padding != null;
    // If padding exists.
    if (hasRangePadding) {
      const { length } = padding;
      const begin = this.#bytesRead;
      if (length > 0 && begin < length) {
        bytesRead = Math.min(size, length - begin);
        const end = begin + bytesRead;
        this.push(padding.subarray(begin, end));
        this.#bytesRead = end;
      }
    }
    // If no padding or read completed.
    if (bytesRead <= 0) {
      this.#bytesRead = 0;
      const readState = this.#readState;
      // Change read state.
      switch (readState) {
        case 0 /* ReadState.PREFIX */:
          this.#readState = 1 /* ReadState.RANGE */;
          this.#readFileRange(fd, range, size);
          break;
        case 2 /* ReadState.SUFFIX */:
          this.#currentRangeIndex++;
          this.#readState = 0 /* ReadState.PREFIX */;
          const nextRange = this.#getRange();
          if (nextRange == null) {
            this.push(null);
          } else {
            this.#readFilePadding(fd, nextRange, size);
          }
          break;
      }
    }
  }
  /**
   * @private
   * @method #readFileRange
   * @param fd The file descriptor.
   * @param range The current range.
   * @param size The number of bytes to read.
   */
  #readFileRange(fd, range, size) {
    this.#reading = true;
    const bytesRead = this.#bytesRead;
    const position = range.offset + bytesRead;
    const buffer = Buffer.allocUnsafeSlow(Math.min(size, range.length - bytesRead));
    // Read file range.
    this.#fs.read(fd, buffer, 0, buffer.length, position, (readError, bytesRead, buffer) => {
      this.#reading = false;
      // Tell ._destroy() that it's safe to close the fd now.
      if (this.destroyed) {
        this.emit(DISPOSE_EVENT, readError);
      } else {
        if (readError != null) {
          this.destroy(readError);
        } else {
          if (bytesRead > 0) {
            if (bytesRead !== buffer.length) {
              // Slow path. Shrink to fit.
              // Copy instead of slice so that we don't retain
              // large backing buffer for small reads.
              const chunk = Buffer.allocUnsafeSlow(bytesRead);
              buffer.copy(chunk, 0, 0, bytesRead);
              buffer = chunk;
            }
            this.push(buffer);
            this.#bytesRead += bytesRead;
          } else {
            this.#bytesRead = 0;
            this.#readState = 2 /* ReadState.SUFFIX */;
            this.#readFilePadding(fd, range, size);
          }
        }
      }
    });
  }
  /**
   * @override
   * @method _read
   * @param size The number of bytes to read.
   */
  _read(size) {
    if (!this.#reading) {
      const fd = this.#fd;
      const range = this.#getRange();
      // If fd or range is null, finish stream.
      if (fd == null || range == null) {
        this.push(null);
      } else {
        const readState = this.#readState;
        // Read bytes from range.
        switch (readState) {
          case 0 /* ReadState.PREFIX */:
          case 2 /* ReadState.SUFFIX */:
            this.#readFilePadding(fd, range, size);
            break;
          case 1 /* ReadState.RANGE */:
            this.#readFileRange(fd, range, size);
            break;
        }
      }
    }
  }
  /**
   * @private
   * @method #dispose
   * @param error The error.
   * @param callback The callback.
   */
  #dispose(error, callback) {
    const fd = this.#fd;
    if (fd != null) {
      this.#fd = null;
      // Close the fd.
      this.#fs.close(fd, closeError => {
        callback(error ?? closeError);
      });
    } else {
      callback(error);
    }
  }
  /**
   * @override
   * @method _destroy
   * @param error The error.
   * @param callback The callback.
   */
  _destroy(error, callback) {
    // Wait I/O completion.
    if (this.#reading) {
      this.once(DISPOSE_EVENT, disposeError => {
        this.#dispose(error ?? disposeError, callback);
      });
    } else {
      this.#dispose(error, callback);
    }
  }
}

export { ReadStream };
